//! # User tokens management

use crate::connections::redis_connection;
use crate::models::{User, UserID};
use crate::utils::string_utils;
use crate::utils::time_utils::time;
use actix_web::dev::Payload;
use actix_web::{FromRequest, HttpRequest};
use std::future::{ready, Ready};
use std::time::Duration;

#[derive(Debug, Clone)]
pub struct LoginTokenValue(String);

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct LoginToken {
    #[serde(rename = "e")]
    expire: u64,
    #[serde(rename = "h")]
    hb: u64,
    #[serde(rename = "u")]
    pub user_id: UserID,
}

impl LoginToken {
    pub fn new(user: &User) -> (String, Self) {
        let key = format!("{}-{}", user.id().0, string_utils::rand_str(40));

        (
            key,
            Self {
                expire: time() + 3600 * 24,
                hb: time(),
                user_id: user.id(),
            },
        )
    }

    pub fn is_expired(&self) -> bool {
        self.expire < time() || self.hb + 3600 < time()
    }

    pub fn refresh_hb(&self) -> Option<Self> {
        if self.hb + 60 * 5 < time() {
            let mut new = self.clone();
            new.hb = time();
            Some(new)
        } else {
            None
        }
    }

    pub fn redis_lifetime(&self) -> Duration {
        Duration::from_secs(if self.expire <= time() {
            0
        } else {
            self.expire - time()
        })
    }
}

/// Get redis key for a given login token
fn redis_key(tok: &str) -> String {
    format!("tok-{tok}")
}

/// Generate a new login token
pub async fn gen_new_token(user: &User) -> anyhow::Result<String> {
    let (key, token) = LoginToken::new(user);
    redis_connection::set_value(&redis_key(&key), &token, token.redis_lifetime()).await?;
    Ok(key)
}

/// Delete a login token
pub async fn delete_token(token: &LoginTokenValue) -> anyhow::Result<()> {
    redis_connection::remove_value(&redis_key(&token.0)).await?;
    Ok(())
}

/// Get a user information from its token
async fn get_token(key: &str) -> anyhow::Result<Option<LoginToken>> {
    let token = match redis_connection::get_value::<LoginToken>(&redis_key(key)).await? {
        None => {
            log::error!("Could not find token for key '{}' (key absent)", key);
            return Ok(None);
        }
        Some(t) => t,
    };

    if token.is_expired() {
        log::error!("Could not find token for key '{}' (token expired)", key);
        return Ok(None);
    }

    // Check if heartbeat must be updated
    if let Some(renew) = token.refresh_hb() {
        redis_connection::set_value(&redis_key(key), &renew, renew.redis_lifetime()).await?;
    }

    Ok(Some(token))
}

impl FromRequest for LoginTokenValue {
    type Error = actix_web::Error;
    type Future = Ready<Result<Self, Self::Error>>;

    fn from_request(req: &HttpRequest, _payload: &mut Payload) -> Self::Future {
        ready(match req.headers().get("X-Auth-Token") {
            Some(v) => Ok(LoginTokenValue(v.to_str().unwrap_or("").to_string())),
            None => Err(actix_web::error::ErrorBadRequest(
                "Missing auth token header!",
            )),
        })
    }
}

impl FromRequest for LoginToken {
    type Error = actix_web::Error;
    type Future = futures_util::future::LocalBoxFuture<'static, Result<Self, Self::Error>>;

    fn from_request(req: &HttpRequest, _payload: &mut Payload) -> Self::Future {
        let req = req.clone();
        Box::pin(async move {
            let token = LoginTokenValue::extract(&req).await?;

            let token = match get_token(&token.0).await {
                Err(e) => {
                    log::error!("Failed to load auth token! {}", e);
                    return Err(actix_web::error::ErrorInternalServerError(
                        "Failed to check auth token!",
                    ));
                }
                Ok(None) => {
                    return Err(actix_web::error::ErrorPreconditionFailed(
                        "Invalid auth token!",
                    ));
                }
                Ok(Some(t)) => t,
            };

            Ok(token)
        })
    }
}
